static char rcsid[] = "@(#)$Id: con_imap.c,v 2.3 2013-10-13 07:18:53 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *                  (was hurtta+elm@posti.FMI.FI)  
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "def_mbox.h"

#ifdef REMOTE_MBX
#include "ss_imp.h"
#include "mbximap_imp.h"

#include "s_me.h"

static const unsigned char *cs2us P_((const char *str));
static const unsigned char *cs2us(str)
     const char *str;
{
    return (const unsigned char *)str;
}

static char *us2s P_((unsigned char *str));
static char *us2s(str)
     unsigned char *str;
{
    return (char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str)
     char *str;
{
    return (unsigned char *)str;
}


DEBUG_VAR(Debug,__FILE__,"imap");

static void  cache_zero_imap P_((struct connection_cache *c));
static void  cache_free_imap P_((struct connection_cache *c));
static int  cache_open_imap P_((struct connection_cache *c));
static int  cache_login_imap P_((struct connection_cache *c,
				 const struct string *password /* May be NULL */));
static int  cache_close_imap P_((struct connection_cache *c));
static void cache_folder_from_imap P_((struct connection_cache *c,
				       struct folder_info *f));
static void cache_browser_from_imap P_((struct connection_cache *c,
					struct folder_browser *d));

struct connection_type IMAP_connection = { 
    CONNECTION_TYPE_magic,
    "IMAP",
    cache_zero_imap,
    cache_free_imap,
    cache_open_imap,
    cache_login_imap,
    cache_close_imap,     
    cache_folder_from_imap,
    cache_browser_from_imap
};

void set_imap_connection_state(Ch,st)
     struct connection_cache  *Ch; 
     imap_states st; 
{
    Ch->a.imap_con->imap_state = st;
}

/* ------------------------------------------------------------------------ */

static void add_token P_((struct Imap_Token_Buffer *X, struct imap_token T));
static void add_token(X,T)
     struct Imap_Token_Buffer *X; 
     struct imap_token T;
{
    X->tokens = safe_realloc(X->tokens,  
			     (X->token_count+1) * sizeof (struct imap_token));
    X->tokens[X->token_count++] = T;
}

static void free_tokens P_((struct Imap_Token_Buffer *X));
static void free_tokens(X)
     struct Imap_Token_Buffer *X;
{

    if (X -> tokens) {
	int i;
	
	for (i = 0; i < X -> token_count; i++) {
	    if (X -> tokens[i].str) {
		free (X -> tokens[i].str);
		X -> tokens[i].str = NULL;
	    }
	    X -> tokens[i].imap_token = imap_zero;
	}
	
	free(X -> tokens);
	X -> tokens = NULL;
	X -> token_count = 0;
    }
}



static struct kwd {
    char         *   kw;
    enum token_type  t;
} keywords[] = {
    { "OK",       imap_status_keyword },
    { "NO",       imap_status_keyword },
    { "BAD",      imap_status_keyword },
    { "PREAUTH",  imap_status_keyword },
    { "BYE",      imap_status_keyword }
};

#define CAPA_IMAP4rev1           1

static struct capab {
    char         *    capa;
    long              mask;
} capabilities[] = {
    { "IMAP4rev1",   CAPA_IMAP4rev1 }
};

struct imap_flag IMAP_flags[] = {
    { "\\Seen",        0,           0,          IMAP_Seen     },
    { "\\Answered",    REPLIED_TO,  0,          IMAP_Answered },
    { "\\Flagged",     0,           S1_FLAGGED, IMAP_Flagged  },
    { "\\Deleted",     DELETED,     0,          IMAP_Deleted  },
    { "\\Draft",       0,           0,          IMAP_Draft  },
    { "\\Recent",      NEW,         0,          IMAP_Recent  },
} ;
int IMAP_flags_count = sizeof IMAP_flags / sizeof (struct imap_flag);

static struct browser_flag {
    char         *   flag;
    int              browser_flag;
} BROWSER_flags[] = {
    { "\\Marked",      BROWSER_MARKED   },
    { "\\Noinferiors", BROWSER_NODIR    },
    { "\\Noselect",    BROWSER_NOFOLDER },
    { "\\Unmarked",    0 },
} ; 

static void add_to_write_buffer P_((struct IMAP_CON * M,
				    char **str));
static void add_to_write_buffer(M,str)
     struct IMAP_CON * M;
     char **str;
{
    int l = strlen(*str);
    
    add_to_Write_Buffer(& (M->write_buffer), str, l);
}

static int imap_gen_writable_data P_((struct IMAP_CON * M));
static int imap_gen_writable_data(M)
     struct IMAP_CON * M;
{
    char * t;

    if (M->write_buffer.write_len > 0) {
	/* Have incomplete data ... */
	return 1;
    }

    free_Write_Buffer(& (M->write_buffer));

    if (M->imap_state != IMAP_command)
	return 0;    /* No data available */

    if (M->token_index >= M->write_tokens.token_count)
	return 0;    /* No data available */

    switch(M->token_index) {
    case -2 :        /* NOOP command */
	M->write_buffer.write_buffer = elm_message (FRM("%s NOOP\r\n"),
				       M->current_NOOP_tag);
	M->write_buffer.write_len    = strlen(M->write_buffer.write_buffer);
	return 1;

    case -1:         /* Normal command */
	M->write_buffer.write_buffer = elm_message(FRM("%s "),
				      M->current_tag);
	M->write_buffer.write_len    = strlen(M->write_buffer.write_buffer);
	M->literal_mode = no_literal;
	M->token_index++;
    }

    while (M->token_index < M->write_tokens.token_count) {
	char *s1 = NULL;
	char * space = "";
	if (M->token_index > 0 &&
	    M->write_tokens.tokens[M->token_index-1].imap_token != 
	    imap_list_begin)
	    space = " ";
	
	switch(M->write_tokens.tokens[M->token_index].imap_token) {
	case imap_literal:
	    switch(M->literal_mode) {
		char *s;
	    case continue_literal:
		if (M->write_buffer.write_buffer != NULL)
 		    panic("MBX PANIC",__FILE__,__LINE__,
			  "imap_gen_writable_data",
			  "write_buffer not NULL",0);
		M->write_buffer.write_buffer = 
		    M->write_tokens.tokens[M->token_index].str;
		M->write_buffer.write_len    = 
		    M->write_tokens.tokens[M->token_index].len_or_value;

		/* So that it is not free()ed twice */
		M->write_tokens.tokens[M->token_index].str = NULL; 

		M->literal_mode = no_literal;
		M->token_index++;
		continue;       /* Write literal */            
	    case waiting_literal_continue:
		DPRINT(Debug,7,(&Debug, 
				"imap_gen_writable_data=0:  Waiting literal continue\n"));
		return 0;
	    case no_literal:
		s = elm_message(FRM("%s{%d}\r\n"),
				space,
				M->write_tokens.tokens[M->token_index].
				len_or_value);
		add_to_write_buffer(M,&s);  /* May free() s */
		M->literal_mode = waiting_literal_continue;
		return 1;
	    }
	    panic("MBX PANIC",__FILE__,__LINE__,
		  "imap_gen_writable_data",
		  "Bad literal mode",0);
	case imap_number:
	    s1 = elm_message(FRM("%s%d"),space,
			     M->write_tokens.tokens[M->token_index].
			     len_or_value);
	    M->token_index++;
	    break;
	case imap_list_begin:
	    s1 = elm_message(FRM("%s("),space);
	    M->token_index++;
	    break;
	case imap_list_end:
	    s1 = elm_message(FRM(")"));
	    M->token_index++;
	    break;
	case imap_atom:
	    s1 = elm_message(FRM("%s%s"),space,
			     M->write_tokens.tokens[M->token_index].str);
	    M->token_index++;
	    break;
	case imap_string:
	    /* NOTE:
	     *
	     * 1) We must NOT convert date-time to literal:
	     *        16-Sep-2000 22:52:00 +0300
	     *
	     * 2) Also it is no effective to convert % to literal
	     *
	     */

	    if (strspn(M->write_tokens.tokens[M->token_index].str,
		       " 1234567890%:+-.!*,()[]#_<>qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM") < 
		M->write_tokens.tokens[M->token_index].len_or_value) {
		DPRINT(Debug,50,(&Debug, 
				"imap_gen_writable_data: %d string \"%s\" converting to literal...\n",
				M->token_index,
				M->write_tokens.tokens [M->token_index].str));
		M->write_tokens.tokens[M->token_index].imap_token =
		    imap_literal;
		break;   /* RETRY */
	    }
	    s1 = elm_message(FRM("%s\"%s\""),space,
			     M->write_tokens.tokens[M->token_index].str);
	    M->token_index++;
	    break;
	default:
	    panic("MBX PANIC",__FILE__,__LINE__,
		  "imap_gen_writable_data",
		  "Bad token type",0);
	}

	if (s1)
	    add_to_write_buffer(M,&s1);  /* May free() s1 */
    }

    /* Write end of command (CRLF) */
    t = safe_strdup("\r\n");
    add_to_write_buffer(M,&t);  /* May free() t */

    return 1;  /* have writable data */
}

/* -1 == failure
   0  == continue   (read more data)
   1  == ready
*/
static int tokenize P_((struct IMAP_CON * M));
static int tokenize(M)
     struct IMAP_CON * M;
{
    int idx;
    int len;
    int ret;

    if (M->wanna_literal) {
	struct imap_token T;

    try_literal:
	if (M->tokenizer_state != state_list)
	    panic("MBX PANIC",__FILE__,__LINE__,"tokenize",
		  "Bad tokenizer_state (IMAP)",0);

	if (M->read_buffer.read_len < M->literal_len) {
	    DPRINT(Debug,7,(&Debug, 
			    "tokenize=0 (incomplete literal)\n"));
	    return 0;
	}

	T.imap_token = imap_literal;
	T.str = safe_malloc(M->literal_len+1);
	memcpy(T.str,M->read_buffer.read_buffer,M->literal_len);
	T.str[M->literal_len] = '\0';
	T.len_or_value = M->literal_len;
	add_token(&(M->read_tokens),T);
	cut_line(&(M->read_buffer),M->literal_len);
	M->wanna_literal = 0;
    }

    idx = 0;
    ret = 1;
    len = find_crlf(&(M->read_buffer),1);
    if (!len) {
	DPRINT(Debug,7,(&Debug, "tokenize=0 (no data)\n"));
	return 0;
    }

    while (idx < len) {	
	struct imap_token T;
	char * p = M->read_buffer.read_buffer +idx;
	int left = len - idx;
	
	if (!*p) {
	    DPRINT(Debug,20,(&Debug, "tokenize: idx=%d   END\n",idx));
	    break;
	}

	T.len_or_value = 0;
	T.str          = NULL;
	T.imap_token   = imap_zero;

	switch (M->tokenizer_state) {
	    char  *n;
	    int i;
	case state_tag:
	    n = memchr(p,' ',left);
	    
	    if (NULL == n || n == p) {
	    fail:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnparseableResponse,
				  "Unparseable response from IMAP server"));

		free_Read_Buffer(&(M->read_buffer));		
		DPRINT(Debug,7,(&Debug, "tokenize=-1 (failure)\n"));
		return -1;
	    }
	    *n = '\0';

	    DPRINT(Debug,20,(&Debug, 
			"tokenize: idx=%d  TAG   '%.*s'\n",
			idx,n-p,p));

	    M->tokenizer_state = state_response;	    
	    if (n-p == 1 && p[0] == '+') {
		T.imap_token = imap_continue;
		M->tokenizer_state = state_text;
	    } else if (n-p == 1 && p[0] == '*')
		T.imap_token = imap_notag;
	    else {
		T.imap_token = imap_atom;
		T.str = safe_malloc(n-p+1);
		memcpy(T.str,p,n-p+1);
		T.len_or_value = n-p; 
	    }
	    add_token(&(M->read_tokens),T);

	    idx = (n - M->read_buffer.read_buffer) +1;
	    break;  
	case state_response:
	    if (p[0] >= '0' && p[0] <= '9') {
		char *c;
		T.len_or_value = strtol(p,&c,10);
		if (*c == ' ') {

		    DPRINT(Debug,20,(&Debug, 
				     "tokenize: idx=%d  RESPONSE (number) '%.*s'\n",
				     idx,c-p,p));
		
		    T.imap_token = imap_number;
		    add_token(&(M->read_tokens),T);
		    idx = (c - M->read_buffer.read_buffer) +1;
		    break;
		}
	    }
	    n = memchr(p,' ',left);
	    if (n == p) 
		goto fail;
	    
	    if (NULL == n)
		n = p + left;
	    else 
		*n = '\0';
		
	    DPRINT(Debug,20,(&Debug, 
			     "tokenize: idx=%d  RESPONSE (keyword) '%.*s'\n",
			     idx,n-p,p));

	    for (i = 0; i < sizeof keywords / sizeof (struct kwd); i++)
		if (0 == istrcmp(keywords[i].kw,p))
		    break;
	    
	    if (i < sizeof keywords / sizeof (struct kwd))
		T.imap_token = keywords[i].t;
	    else
		T.imap_token = imap_other_keyword;
	    
	    if (imap_status_keyword == T.imap_token)
		M->tokenizer_state = state_text;
	    else if (imap_other_keyword == T.imap_token)
		M->tokenizer_state = state_list;

	    T.str = safe_malloc(n-p+1);
	    memcpy(T.str,p,n-p+1);
	    T.len_or_value = n-p; 
	    add_token(&(M->read_tokens),T);

	    idx = (n - M->read_buffer.read_buffer) +1;
	    break;
	case state_text:
	    if ('[' == p[0]) {

		DPRINT(Debug,20,(&Debug, 
			    "tokenize: idx=%d  TEXT   [\n",
			    idx));
		
		T.imap_token = imap_code_begin;
		M->tokenizer_state = state_text_code;
		idx++;
		add_token(&(M->read_tokens),T);
		break;
	    }	    

	    DPRINT(Debug,20,(&Debug, 
			     "tokenize: idx=%d  TEXT   '%.*s'\n",
			     idx,left,p));

	    T.imap_token = imap_string;
	    T.str = safe_malloc(left);
	    memcpy(T.str,p,left);
	    T.len_or_value = left-2;   /* Not find_crlf() */

	    add_token(&(M->read_tokens),T);

	    idx += left;
	    break;
	case state_text_code:
	    if (']' == p[0]) {

		DPRINT(Debug,20,(&Debug, 
			    "tokenize: idx=%d  CODE   ]\n",
			    idx));

		T.imap_token = imap_code_end;
		M->tokenizer_state = state_text;
		idx++;
		if (' ' == p[1])   /* Skip separated space between code */
		    idx++;         /* argument and textual code         */
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    /* FALLTHRU */
	case state_list:
	    if ('(' == p[0]) {

		DPRINT(Debug,20,(&Debug, 
			    "tokenize: idx=%d  LIST   (\n",
			    idx));
		
		T.imap_token = imap_list_begin;
		idx++;
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    if (')' == p[0]) {

		DPRINT(Debug,20,(&Debug, 
			    "tokenize: idx=%d  LIST   )\n",
			    idx));

		T.imap_token = imap_list_end;
		idx++;
		if (' ' == p[1])   /* Skip separated space between list */
		    idx++;         /* argument and next argument        */
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    if (p[0] >= '0' && p[0] <= '9') {
		char *c;
		T.len_or_value = strtol(p,&c,10);
		if (*c == ' ' || *c == ')' || *c == ']') {
		    
		    DPRINT(Debug,20,(&Debug, 
				     "tokenize: idx=%d  LIST (number)   '%.*s'\n",
				     idx,c-p,p));
		    
		    T.imap_token = imap_number;
		    add_token(&(M->read_tokens),T);
		    if (*c == ' ')
			idx = (c - M->read_buffer.read_buffer) +1;
		    else
			idx = (c - M->read_buffer.read_buffer);
		    break;
		}
	    }

	    if (M->tokenizer_state != state_text_code) {
		if ('"' == p[0]) {
		    char * x, *y;
		    T.imap_token = imap_string;
		    T.str = safe_malloc(left);
		    y = T.str;

		    for (x = p+1; x < p + left && '\0' != *x && '"' != *x; 
			 x++) {
			if ('\\' == *x)
			    x++;
			if (y < T.str + left -1)
			    *y++ = *x;
		    }
		    if (*x != '"')
			goto fail;
		    x++;
		    *y = '\0';

		    DPRINT(Debug,20,(&Debug, 
				     "tokenize: idx=%d  LIST (string)   '%.*s'\n",
				     idx,y-T.str,T.str));

		    T.len_or_value = y-T.str; 
		    add_token(&(M->read_tokens),T);
		    
		    if (' ' == *x) /* Skip separated space between string */
			x++;       /* argument and next argument          */
		    idx = x - M->read_buffer.read_buffer;		
		    break;
		}
		if (p[0] == '{' && left > 4 && p[left-3] == '}' &&
		    M->tokenizer_state == state_list) {
		    char *c;
		    M->literal_len = strtol(p+1,&c,10);
		    if (c == &(p[left-3])) {   /* Start parsing literal */

			DPRINT(Debug,20,(&Debug, 
					 "tokenize: idx=%d  LIST (literal)   len=%d\n",
					 idx,M->literal_len));
			
			M->wanna_literal = 1;
			idx += left;
			ret = 0;               /* Continue reading ... */
			break;
		    }
		}
	    }

	    if (' ' == p[0]) {
		/* Communigate Pro 3.3.1 produces extra space between
		   CAPABILITY response tokens before auth capabilities
		*/
		DPRINT(Debug,3,(&Debug, 
				"* Format error -- extra space between tokens on IMAP response:\n"));
		DPRINT(Debug,3,(&Debug, 
				">>> %s\n",
				p));
		idx++;
		continue;
	    }


	    T.imap_token = imap_atom;
	    if (M->tokenizer_state == state_text_code)
		i = strcspn(p," ()]");
	    else
		i = strcspn(p,"(){ *%\"");
	    if (i == 0) {
		DPRINT(Debug,9,(&Debug, 
				"tokenize: Possible internal error:'%s'\n",
				p));
		goto fail;
	    }

	    DPRINT(Debug,20,(&Debug, 
			     "tokenize: idx=%d  LIST (atom)   '%.*s'\n",
			     idx,i,p));
	    
	    T.imap_token = imap_atom;
	    T.str = safe_malloc(i+1);
	    memcpy(T.str,p,i);
	    T.str[i] = '\0';
	    idx += i;
	    if (' ' == M->read_buffer.read_buffer[idx])
		idx++;
	    add_token(&(M->read_tokens),T);
	    break;
	default:
	    panic("MBX PANIC",__FILE__,__LINE__,"tokenize",
		  "Bad tokenizer_state (IMAP)",0);
	}
    }
    
    if (!M->wanna_literal) {
	/* Again beginning of line */
	M->tokenizer_state = state_tag;
    }
    cut_line(&(M->read_buffer),len);

    if (M->wanna_literal && M->read_buffer.read_len)
	goto try_literal;

    DPRINT(Debug,7,(&Debug, "tokenize=%d (token count %d)\n",
		    ret,M->read_tokens.token_count));

    return ret;   /* Response parsed? */
}

static void pick_text P_((struct IMAP_CON * M, struct string **Text,
			  char **code));
static void pick_text(M,Text,code)
     struct IMAP_CON * M;
     struct string **Text;
     char **code;
{
    int i;
    int q = 0;
    
    *Text = NULL;
    *code = NULL;

    for (i = 2; i < M->read_tokens.token_count; i++) {
	if (imap_code_begin == M->read_tokens.tokens[i].imap_token)
	    q++;
	else if (imap_code_end == M->read_tokens.tokens[i].imap_token)
	    q--;
	else if (q && imap_atom == M->read_tokens.tokens[i].imap_token &&
		 !*code)
	    *code = M->read_tokens.tokens[i].str;
	else if (!q && imap_string == M->read_tokens.tokens[i].imap_token &&
		 !*Text)
	    /* User readable code may use MIME encoded words on message */
	    *Text = hdr_to_string(HDR_TEXT,
				  M->read_tokens.tokens[i].str,
				  imap_charset,1);	    
    }
}

static int scan_list P_((struct imap_token ** ptr,
			 int                * left,
			 struct imap_token ** found_list,
			 int                * list_len));
static int scan_list(ptr,left,found_list,list_len)
     struct imap_token ** ptr;
     int                * left;
     struct imap_token ** found_list;
     int                * list_len;
{
    struct imap_token * p = *ptr;
    int                 l = *left;
    int i,q=0;

    *found_list = NULL;
    *list_len   = 0;

    if (l < 2 || imap_list_begin != p[0].imap_token)
	return 0;

    *found_list = p+1;
    
    for (i = 0; i < l; i++) {
	if (imap_list_begin == p[i].imap_token)
	    q++;
	else if (imap_list_end == p[i].imap_token) {
	    q--;
	    if (0 == q) {
		*list_len = i-1;
		*ptr      = p+  (i +1);
		*left     = l - (i +1);
		return 1;
	    }
	}
    }
    return 0;
}

void imap_free_reference(item)
     struct imap_reference *item;
{
    item->uid_number        = -1;
    item->imap_flags        = 0;
    item->rfc822_size       = 0;
    
    if (item->internaldate) {
	free(item->internaldate);
	item->internaldate  = NULL;
    }

    if (item->header) {
	free(item->header);
	item->header      = NULL;
	item->header_len = 0;	    
    }
    if (item->body) {
	free(item->body);
	item->body      = NULL;
	item->body_len  = 0;    
    }
}

static void zero_reference P_((struct imap_reference *item));
static void zero_reference(item)
     struct imap_reference *item;
{
    item->uid_number        = -1;
    item->imap_flags        = 0;
    item->rfc822_size       = 0;
    item->internaldate      = NULL;
    item->header_len        = 0;
    item->header            = 0;
    item->body_len          = 0;
    item->body              = 00;
}


static int got_header P_((struct imap_reference * item,
			  struct imap_token * lst));
static int got_header(item,lst)
     struct imap_reference * item;
     struct imap_token * lst;
{
    /* Originally malloced by tokenize() */
    if (item->header) {
	free(item->header);
	item->header = NULL;
    }

    switch(lst->imap_token) {
    case imap_atom:
	if (0 == istrcmp(lst->str,"NIL")) {
	    item->header     = NULL;
	    item->header_len = 0;
	    return 1;       /* OK */
	}
	return 0;           /* Bad */
    case imap_string:
    case imap_literal:
	/* Move malloced data to struct imap_reference */
	item->header     = lst->str;
	item->header_len = lst->len_or_value;
	lst->str         = NULL;  /* So that free_tokens() does not try free 
				   * it  
				   */
	return 1;           /* OK */
    default:
	return 0;   /* Bad */
    }
}

static int got_body P_((struct imap_reference * item,
			struct imap_token * lst));
static int got_body(item,lst)
     struct imap_reference * item;
     struct imap_token * lst;
{
    /* Originally malloced by tokenize() */
    if (item->body) {
	free(item->body);
	item->body = NULL;
    }

    switch(lst->imap_token) {
    case imap_atom:
	if (0 == istrcmp(lst->str,"NIL")) {
	    item->body     = NULL;
	    item->body_len = 0;
	    return 1;       /* OK */
	}
	return 0;           /* Bad */
    case imap_string:
    case imap_literal:
	/* Move malloced data to struct imap_reference */
	item->body     = lst->str;
	item->body_len = lst->len_or_value;
	lst->str       = NULL;  /* So that free_tokens() does not try free 
				 * it  
				 */
	return 1;           /* OK */
    default:
	return 0;   /* Bad */
    }

}

static void process_list_flags P_((struct imap_dir_entry * item,
				   struct imap_token * lst,
				   int                 lst_len));
static void process_list_flags(item,lst,lst_len)
     struct imap_dir_entry * item;
     struct imap_token     * lst;
     int                     lst_len;
{
    DPRINT(Debug,9,(&Debug, 
		    "process_list_flags: len=%d",
		    lst_len));
    while (lst_len > 0) {
	struct imap_token * kwd = lst;
	lst_len--;
	lst++;
	if (imap_atom == kwd->imap_token) {
	    int j;
	    int found = 0;
	    for (j = 0; 
		 j < sizeof BROWSER_flags / 
		     sizeof (struct browser_flag);
		 j++) {
		if (0 == istrcmp(kwd->str,
				 BROWSER_flags[j].flag)) {
		    item->flags |= BROWSER_flags[j].browser_flag;
		    found++;
		    break;
		}
	    }
	    if (!found) {
		DPRINT(Debug,9,(&Debug, 
				" [Unknown flag %s] ",
				kwd->str));
	    }
	    DPRINT(Debug,10,(&Debug, 

			     " [known flags:"));
	    for (j = 0;
		 j < sizeof BROWSER_flags / sizeof (struct browser_flag);
		 j++) {
		if (BROWSER_flags[j].browser_flag &&
		    (item->flags & BROWSER_flags[j].browser_flag)
		    == BROWSER_flags[j].browser_flag)
		    DPRINT(Debug,10,(&Debug, 
				     " %s",BROWSER_flags[j].flag));
	    }
	    DPRINT(Debug,9,(&Debug, 
			    "]"));
	} else {
	    DPRINT(Debug,9,(&Debug, 
			    " ???\n"));
	    DPRINT(Debug,9,(&Debug, 
			    "process_list_flags: List response  (..) flag item have token type %d\n",
			    kwd->imap_token));
	    break;	/* Quit processing */	
	}
    }
    DPRINT(Debug,9,(&Debug, "\n"));
}


static void process_fetch_items P_((struct imap_reference * item,
				    struct imap_token * lst,
				    int                 lst_len));
static void process_fetch_items(item,lst,lst_len)
     struct imap_reference * item;
     struct imap_token     * lst;
     int                     lst_len;
{			
    DPRINT(Debug,7,(&Debug,  
		    "process_fetch_items: len=%d",
		    lst_len));
    while (lst_len > 1) {
	struct imap_token * kwd = lst;
	lst_len--;
	lst++;
	
	if (imap_atom == kwd->imap_token) {
	    if (0 == istrcmp(kwd->str,"UID")) {
		if (imap_number == lst->imap_token) {
		    item->uid_number = lst->len_or_value;

		    DPRINT(Debug,9,(&Debug, 
				    " UID %d",item->uid_number));

		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " UID ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after UID number expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"INTERNALDATE")) {
		if (imap_string == lst->imap_token) {
		    DPRINT(Debug,9,(&Debug,  " INTERNALDATE %s (len %d)",
				    lst->str,lst->len_or_value));
		    if (lst->len_or_value != 26) {
			DPRINT(Debug,9,(&Debug, "\n"));
			DPRINT(Debug,1,(&Debug, 
				   "process_fetch_items: Fetch response  (..) after INTERNALDATE 26 char string expected, got %d chars\n",
					lst->len_or_value));
		    } else {
			if (item->internaldate)
			    free(item->internaldate);
			item->internaldate = lst->str;
			lst->str          = NULL;   /* Avoid double free() */
		    }
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " INTERNALDATE ????\n"));
		    DPRINT(Debug,1,(&Debug, 
			       "process_fetch_items: Fetch response  (..) after INTERNALDATE string expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.SIZE")) {
		if (imap_number == lst->imap_token) {
		    item->rfc822_size = lst->len_or_value;

		    DPRINT(Debug,9,(&Debug, " RFC822.SIZE %d",
				    item->rfc822_size));
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " RFC822.SIZE ????\n"));
		    DPRINT(Debug,9,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after RFC822.SIZE number expected, got type %d\n",
				    lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.HEADER")) {
		if (got_header(item,lst)) {		    
		    DPRINT(Debug,9,(&Debug, " RFC822.HEADER ..."));
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " RFC822.HEADER ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after RFC822.HEADER nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.TEXT")) {
		if (got_body(item,lst)) {		    
		    DPRINT(Debug,9,(&Debug, " RFC822.TEXT ..."));
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " RFC822.TEXT ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after RFC822.BODY nstring expected, got type %d\n",
				    lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"BODY[HEADER]")) {
		if (got_header(item,lst)) {		    
		    DPRINT(Debug,9,(&Debug, " BODY[HEADER] ..."));
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " BODY[HEADER] ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after BODY[HEADER] nstring expected, got type %d\n",
				    lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"BODY[TEXT]")) {
		if (got_body(item,lst)) {		    
		    DPRINT(Debug,9,(&Debug, " BODY[TEXT] ..."));
		    lst_len--;
		    lst++;
		} else {
		    DPRINT(Debug,9,(&Debug, " BODY[TEXT] ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after BODY[TEXT] nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"FLAGS")) {
		struct imap_token * flg;
		int                 flg_len;
		
		item->imap_flags = 0;
		
		DPRINT(Debug,9,(&Debug, " FLAGS "));
		if (scan_list(&lst,&lst_len,&flg,&flg_len)) {
		    int i, j;
		    
		    DPRINT(Debug,9,(&Debug, "(listlen=%d)",flg_len));
		    for (i = 0; i < flg_len; i++) {
			if (imap_atom == flg[i].imap_token) {
			    int found = 0;
			    for (j = 0; 
				 j < sizeof IMAP_flags / 
				     sizeof (struct imap_flag);
				 j++) {
				if (0 == istrcmp(flg[i].str,
						 IMAP_flags[j].flag)) {
				    item->imap_flags |= 
					IMAP_flags[j].mask_imap;
				    found++;
				    break;
				}
			    }
			    if (!found) {
				DPRINT(Debug,9,(&Debug, 
						" [Unknown flag %s] ",
						flg[i].str));
			    }
			}			
		    }
		    DPRINT(Debug,10,(&Debug, 
				     " [known flags:"));
		    for (j = 0;
			 j < sizeof IMAP_flags / sizeof (struct imap_flag);
			 j++) {
			if ((item->imap_flags & IMAP_flags[j].mask_imap)
			    == IMAP_flags[j].mask_imap)
			    DPRINT(Debug,10,(&Debug, " %s",IMAP_flags[j].flag));
		    }
		    DPRINT(Debug,10,(&Debug, "]"));
		} else {
		    DPRINT(Debug,9,(&Debug, " ???\n"));
		    DPRINT(Debug,1,(&Debug, 
				    "process_fetch_items: Fetch response  (..) after FLAGS (list) expected, got type %d\n",
				    lst->imap_token));
		    break; /* Quit processing */
		}
	    } else {
		DPRINT(Debug,9,(&Debug, " ???\n"));
		DPRINT(Debug,1,(&Debug, 
				"process_fetch_items: Fetch response  (..) list item %s unsupported\n",
				kwd->str));
		break; /* Quit processing */
	    }				    
	} else {
	    DPRINT(Debug,9,(&Debug, " ???\n"));
	    DPRINT(Debug,1,(&Debug, 
			    "process_fetch_items: Fetch response  (..) list item have token type %d\n",
			    kwd->imap_token));
	    break;	/* Quit processing */	
	}
    }    
    DPRINT(Debug,7,(&Debug, "\n"));
}

static int imap_read_stream P_((struct streamsched *ss,void * data));

void imap_clear_command(con)
     struct connection_cache *con;
{
    struct IMAP_CON * const M = con->a.imap_con;

    DPRINT(Debug,7,(&Debug, 
		"imap_clear_command: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));

    if (!M) {
	DPRINT(Debug,7,(&Debug,
			"imap_clear_command: NO IMAP_CON\n!"));
	return;
    }

    if (IMAP_idle == M->imap_state) {
	DPRINT(Debug,7,(&Debug, 
		    "imap_clear_command: Already clear!\n"));
    } else if (IMAP_closing == M->imap_state) {
	DPRINT(Debug,7,(&Debug, 
			"imap_clear_command: On closing state!\n"));
    } else if (IMAP_command_ready_NO  != M->imap_state &&
	       IMAP_command_ready_BAD != M->imap_state &&
	       IMAP_command_ready_OK  != M->imap_state) {
	DPRINT(Debug,7,(&Debug, 
			"imap_clear_command: WRONG STATE!\n"));
	M->imap_state = IMAP_error;
    } else
	M->imap_state = IMAP_idle;

    if (M->current_NOOP_tag &&
	M->imap_state != IMAP_error) {
	DPRINT(Debug,3,(&Debug, 
		    "imap_clear_command: Waiting for idle NOOP complete...\n"));
	while(M->current_NOOP_tag && 
	      M->imap_state != IMAP_error &&
	      con->C.stream) {
	    WaitStreamFor(con->C.stream,SS_read_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}

	DPRINT(Debug,3,(&Debug, 
			"imap_clear_command: idle NOOP wait completed\n"));
    }
    
    /* Reset possible incomplete send literal */
    free_tokens(& (M->write_tokens));
    M->literal_mode = no_literal;
    M->token_index  = -1;

    free_Write_Buffer(& (M->write_buffer));
    
    if (M->current_tag) {
	free(M->current_tag);
	M->current_tag = NULL;
    }

    if (M->command_result_code) {
	free(M->command_result_code);
	M->command_result_code = NULL;
    }
}    

#define IMAP_TIMEOUT  25

struct string *conv_from_imap_name(name)
     const char *name;
{
    struct string *ret = NULL;
    int L;

    /* Try IMAP encoding and if it fails, fall back to imap_charset.
       Do not return IMAP encoding is string's charset, because
       there is danger that it can leak out. Instead return string
       with well known encoding of UNICODE (ie. UTF-7)
    */
       

    if (IMAP_name_convention && name[0] && 0 < (L = strlen(name))) {
	struct charset_state * P = new_state(IMAP_ENCODING);
	uint16 * vector = safe_malloc(L * sizeof (vector[0]));
	const char *s;
	int X = 0;
	int ok = 1;
	charset_t utf7 = NULL;

	DPRINT(Debug,12,(&Debug,  "conv_from_imap_name: %s =",name));
	for (s = name; *s && X < L; s++) {
	    if (!add_streambyte_to_state(P, (unsigned char)*s)) {
		DPRINT(Debug,12,(&Debug,  " FAIL"));
		ok = 0;
		break;
	    }
	    if (state_ready(P)) {
		vector[X++] = give_unicode_from_state(P);
		DPRINT(Debug,12,(&Debug,  " %04X",vector[X-1]));
		reset_state(P,0);
	    }
	}
	DPRINT(Debug,12,(&Debug,  "\n"));

	if (!ok) {
	    DPRINT(Debug,7,(&Debug, 
			    "conv_from_imap_name: Not IMAP encoded: %s\n",
			    name));
	} else if (NULL != (utf7 = MIME_name_to_charset("UTF-7",0))) {
	    ret = new_string(utf7);
	    add_unicode_to_string(ret,X,vector);
	    DPRINT(Debug,7,(&Debug, 
			    "conv_from_imap_name: IMAP encoded, len=%d: %s (%S)\n",
			    X,name,ret));
	}

	free_state(&P);
	free(vector);
    }


    if (!ret)
	ret = new_string2(imap_charset,cs2us(name));

    {
	const char * MIME_name UNUSED_VAROK = get_string_MIME_name(ret);
        const charset_t cs     UNUSED_VAROK = get_string_type(ret);

	DPRINT(Debug,7,(&Debug, 
			"conv_from_imap_name=%S [type=%p '%s']   (name=%s) \n",
			ret,
                        cs,
                        MIME_name ? MIME_name : "<no MIME name>",
                        name));

    }
    return ret;
}


/* Special return value 2 indivates that caller should re-enable 
 * imap_write_stream()
 */
static int parse_response P_((struct connection_cache *con));
static int parse_response(con)
     struct connection_cache *con;

{
    struct IMAP_CON     * M_a = con->a.imap_con;
    struct IMAP_MBX     * M_b = NULL;
    struct IMAP_BROWSER * M_c = NULL;

    if (con->f) {
	if (con->f->folder_type != &imap_mbx) 
	    panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
		  "Bad folder type attached to connection",0);
	else
	    M_b = &(con->f->p->a.imap_mbx);
    }

    if (con->d) {
	if (con->d->type != &imap_browser)
	    panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
		  "Bad browser type attached to connection",0);
	else {
	    M_c = con->d->a.imap_browser;
    
	    if (IMAP_BROWSER_magic != M_c->magic)
		panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
		      "Bad magic type",0);
	}
    }    

    if (!M_a) {
	panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
	      "No imap connection",0);
    }

    if (M_a->read_tokens.token_count < 2) {
	DPRINT(Debug,1,(&Debug, "parse_response: token_count=%d\n",
		   M_a->read_tokens.token_count));
	lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			  "Bad response from IMAP server"));
	return 0;
    }

    switch (M_a->read_tokens.tokens[0].imap_token) {
    case imap_notag:
	switch (M_a->read_tokens.tokens[1].imap_token) {
	    struct string *Text;
	    char *code;   /* Do not free() this -- this is just pointer
			     to read_tokens area
			  */
	case imap_status_keyword:
	    Text = NULL;
	    code = NULL;
	    pick_text(M_a,&Text,&code);
	    if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"BYE")) {
		M_a->imap_state = IMAP_closing;
		if (Text)
		    lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapBye,
					  "IMAP server: %S"),
				  Text);		
	    } else if (code && Text && 0 == istrcmp(code,"ALERT")) {
		lib_error(CATGETS(elm_msg_cat, MeSet,MeImapAlert,
				  "IMAP alert: %S"),
			  Text);		
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"OK")) {
		if (IMAP_show_greeting && !code)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapServer,
				      "IMAP server: %S"),
			      Text);		
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"NO")) {
		if (IMAP_show_greeting)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapWarning,
				      "IMAP warning: %S"),
			      Text);		
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"BAD")) {
		if (IMAP_show_error)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapProtoError,
				      "IMAP error: %S"),
			      Text);			    
	    }
	    if (Text)
		free_string(&Text);
	    break;
	case imap_other_keyword:

	    if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"CAPABILITY")) {
		int i,j;
		M_a->capability_bits = 0;
		
#ifdef USE_DLOPEN
		free_only_imap_capa_libs( & (M_a->capability_libs),
					  & (M_a->capability_lib_count));
#endif		

		for (i = 2; i < M_a->read_tokens.token_count; i++) {
		    if (imap_atom == M_a->read_tokens.tokens[i].imap_token) {
			
			DPRINT(Debug,2,(&Debug,  
					"IMAP server capacity: '%s'\n",
					M_a->read_tokens.tokens[i].str));

#ifdef USE_DLOPEN
		    probe_imap_capa_lib( & (M_a->capability_libs),
					 & (M_a->capability_lib_count),
					 M_a->read_tokens.tokens[i].str);
#endif

			for (j = 0; 
			     j < sizeof capabilities / sizeof (struct capab);
			     j++) {
			    if (0 == istrcmp(M_a->read_tokens.tokens[i].str,
					     capabilities[j].capa)) 
				M_a->capability_bits |= capabilities[j].mask;
			}
		    }
		}
		DPRINT(Debug,9,(&Debug, 
				"parse_response: (known) Capabilities:"));
		for (j = 0;
		     j < sizeof capabilities / sizeof (struct capab);
		     j++) {
		    if ((M_a->capability_bits & capabilities[j].mask)
			== capabilities[j].mask)
			DPRINT(Debug,9,(&Debug, " %s",capabilities[j].capa));
		}
		DPRINT(Debug,9,(&Debug, "\n"));

	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"LIST")) {
		if (M_c) {
		    struct imap_token * ptr  = 
			M_a->read_tokens.tokens+2;
		    int                left = 
			M_a->read_tokens.token_count-2;
		    struct imap_token * lst;
		    int                 lst_len;


		    M_c->dir_entries = 
			safe_realloc(M_c->dir_entries,
				     (M_c->dir_entry_count +1) *
				     sizeof (struct imap_dir_entry));

		    M_c->dir_entries[M_c->dir_entry_count].flags = 0;
		    if (!scan_list(&ptr,&left,&lst,&lst_len) ||
			left != 2) {
			DPRINT(Debug,1,(&Debug, 
				   "parse_response: scan_list failed to parse LIST response list, left=%d\n",
				   left));
			lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
					  "Bad response from IMAP server"));
			return 0;
		    } else {
			process_list_flags(M_c->dir_entries+
					   M_c->dir_entry_count,
					   lst,lst_len);
			if (imap_atom == ptr[0].imap_token &&
			    0 == istrcmp(ptr[0].str,"nil"))
			    M_c->dir_entries[M_c->dir_entry_count].sep ='\0';
			else if (imap_string == ptr[0].imap_token &&
				 1 == ptr[0].len_or_value)
			    M_c->dir_entries[M_c->dir_entry_count].sep = 
				ptr[0].str[0];
			else {
			    DPRINT(Debug,1,(&Debug, 
				       "parse_response: LIST: Bad separator, type=%d\n",
				       ptr[0].imap_token));
			    lib_error(CATGETS(elm_msg_cat, MeSet,
					      MeUnexpextedResponseImap,
					      "Bad response from IMAP server"));
			    return 0;			    
			}
			
			if (imap_string == ptr[1].imap_token)
			    M_c->dir_entries[M_c->dir_entry_count].
				imap_name = safe_strdup(ptr[1].str);
			else if (imap_literal == ptr[1].imap_token)
			    M_c->dir_entries[M_c->dir_entry_count].
				imap_name = safe_strdup(ptr[1].str);
			else if (imap_atom == ptr[1].imap_token)
			    M_c->dir_entries[M_c->dir_entry_count].
				imap_name = safe_strdup(ptr[1].str);
			else {
			    DPRINT(Debug,1,(&Debug, 
				       "parse_response: LIST: Bad mailbox, type=%d\n",
				       ptr[0].imap_token));
			    lib_error(CATGETS(elm_msg_cat, MeSet,
					      MeUnexpextedResponseImap,
					      "Bad response from IMAP server"));
			    return 0;			    
			}
			M_c->dir_entries[M_c->dir_entry_count].
			    translated_name = 
			    conv_from_imap_name(M_c->dir_entries
						[M_c->dir_entry_count].
						imap_name);
			DPRINT(Debug,30,(&Debug, 
					 "parse_response: %d: list item '%s' separator=%c\n",
					 M_c->dir_entry_count,
					 M_c->dir_entries[M_c->dir_entry_count].
					 imap_name ,
					 M_c->dir_entries[M_c->dir_entry_count].
					 sep));
				  
			DPRINT(Debug,30,(&Debug, 
					 "parse_response: %d: %p  '%S'\n",
					 M_c->dir_entry_count,
					 M_c->dir_entries[M_c->dir_entry_count].
					 translated_name,
					 M_c->dir_entries[M_c->dir_entry_count].
					 translated_name));
			       
			M_c->dir_entry_count++;
		    }
		} else {
		    DPRINT(Debug,1,(&Debug, 
				    "parse_response: Unexpected LIST response\n"));
		}
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"LSUB")) {
		/* TODO: Handle LSUB */
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"STATUS")) {
		/* TODO: Handle STATUS */
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"SEARCH")) {
		/* TODO: Handle STATUS */
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"FLAGS")) {
		int i,j;

		if (M_b) {
		    M_b->flag_bits = 0;

		    for (i = 2; i < M_a->read_tokens.token_count; i++) {
			int found = 0;
			if (imap_atom == M_a->
			    read_tokens.tokens[i].imap_token) {
			    for (j = 0; 
				 j < sizeof IMAP_flags / 
				     sizeof (struct imap_flag);
				 j++) {
				if (0 == istrcmp(M_a->
						 read_tokens.tokens[i].str,
						 IMAP_flags[j].flag)) {
				    M_b->flag_bits |= IMAP_flags[j].mask_imap;
				    found++;
				    break;
				}			
			    }
			    if (!found) {
				DPRINT(Debug,9,(&Debug, 
						"parse_response: Unknown flag %s\n",
						M_a->read_tokens.tokens[i].str));
			    }
			}
		    }
		    DPRINT(Debug,9,(&Debug, 
				"parse_response: (known) Flags:"));
		    for (j = 0;
			 j < sizeof IMAP_flags / sizeof (struct imap_flag);
			 j++) {
			if ((M_b->flag_bits & IMAP_flags[j].mask_imap)
			    == IMAP_flags[j].mask_imap)
			    DPRINT(Debug,9,(&Debug, " %s",IMAP_flags[j].flag));
		    }
		    DPRINT(Debug,9,(&Debug, "\n"));
		} else {
		    DPRINT(Debug,1,(&Debug, 
				    "parse_response: Unexpected FLAGS response\n"));
		}
	    }
	    break;
	case imap_number:
	    if (imap_other_keyword == M_a->read_tokens.tokens[2].imap_token) {

		if (0 == istrcmp(M_a->read_tokens.tokens[2].str,"RECENT")) {
		    if (M_b)
			M_b->num_recent = 
			    M_a->read_tokens.tokens[1].len_or_value;
		    else {
			DPRINT(Debug,1,(&Debug, 
					"parse_response: Unexpected RECENT response\n"));
		    }
		} else  if (0 == istrcmp(M_a->read_tokens.tokens[2].str,
					 "EXISTS")) {
		    int val = M_a->read_tokens.tokens[1].len_or_value ;

		    DPRINT(Debug,9,(&Debug, 
				    "parse_response: (mailbox index %d) EXISTS\n",
				    val));

		    if (M_b) {
			/*
			  M_b->reference_count == number of messages + 1
			  M_b->references [0]     is not used
			*/
		    
			if (val+1 < M_b->reference_count) {
			    DPRINT(Debug,1,(&Debug, 
				       "parse_response: Bad EXISTS value: %d\n",
				       val));
			} else if (val+1 > M_b->reference_count) {
			    DPRINT(Debug,10,(&Debug, 
					     "parse_response: %d mails\n",
					     val));
			    M_b->references = safe_realloc(M_b->references,
							   (val +1) *
							   sizeof (struct 
								   imap_reference));
			
			    while (M_b->reference_count < val+1) {
				struct imap_reference *item = 
				    &(M_b->references[M_b->reference_count]);
				zero_reference (item);
				M_b->reference_count++;
			    }
			}
		    } else {
			DPRINT(Debug,1,(&Debug, 
					"parse_response: Unexpected RECENT response\n"));
		    }
		} else  if (0 == istrcmp(M_a->read_tokens.tokens[2].str,
					 "EXPUNGE")) {
		    int val = M_a->read_tokens.tokens[1].len_or_value ;

		    DPRINT(Debug,9,(&Debug, 
				    "parse_response: (mailbox index %d) EXPUNGE\n",
				    val));

		    if (M_b) {

			/* M_b->references[0] is not used ...
			   M_b->references[ 1 .. num of messages ] is on use
			   
			   M_b->reference_count == num of messages +1
			*/

			if (val < 0 || val >= M_b->reference_count) {
			    DPRINT(Debug,1,(&Debug, 
					    "parse_response: Bad EXPUNGE value: %d\n",
					    val));
			} else {
			    int i;
			    imap_free_reference (& (M_b->references[val]));

			    for (i = val+1; i < M_b->reference_count; i++) {
				M_b->references[i-1] = M_b->references[i];
			    }
			    M_b->reference_count--;			
			}
		    } else {
			DPRINT(Debug,1,(&Debug, 		   
					"parse_response: Unexpected EXPUNGE response\n"));
		    }
		} else  if (0 == istrcmp(M_a->read_tokens.tokens[2].str,
					 "FETCH")) {
		    int val = M_a->read_tokens.tokens[1].len_or_value ;

		    DPRINT(Debug,9,(&Debug, 
				    "parse_response: (mailbox index %d) FETCH\n",val));

		    if (M_b) {
			/* M_b->references[0] is not used ...
			   M_b->references[ 1 .. num of messages ] is on use
		       
			   M_b->reference_count == num of messages +1
			*/

			if (val < 0 || val >= M_b->reference_count) {
			    DPRINT(Debug,1,(&Debug, "parse_response: Bad FETCH value: %d\n",
					    val));
			} else {
			    struct imap_token * ptr  = 
				M_a->read_tokens.tokens+3;
			    int                left = 
				M_a->read_tokens.token_count-3;
			    struct imap_token * lst;
			    int                 lst_len;
			
			    if (!scan_list(&ptr,&left,&lst,&lst_len) ||
				left > 0) {
				DPRINT(Debug,1,(&Debug, 
						"parse_response: scan_list failed to parse FETCH response list\n"));
				lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
						  "Bad response from IMAP server"));
				return 0;
			    } else {
				process_fetch_items(M_b->references+val,
						    lst,lst_len);
			    }
			}
		    } else {
			DPRINT(Debug,1,(&Debug, 
					"parse_response: Unexpected FETCH response\n"));
		    }
		} else {
		    DPRINT(Debug,1,(&Debug, 
				    "parse_response: Unsupported data %s\n",
				    M_a->read_tokens.tokens[2].str));
		}
	    } else {
		DPRINT(Debug,1,(&Debug, 
				"parse_response: token[2] type =%d\n",
				M_a->read_tokens.tokens[2].imap_token));
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
				  "Bad response from IMAP server"));
		return 0;
	    }
	    break;
	default:
	    DPRINT(Debug,1,(&Debug, 
			    "parse_response: token[1] type =%d\n",
			    M_a->read_tokens.tokens[1].imap_token));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	break;
    case imap_atom:
	if (M_a->current_NOOP_tag &&
	    0 == strcmp(M_a->current_NOOP_tag,
			M_a->read_tokens.tokens[0].str)) {
	    free(M_a->current_NOOP_tag);
	    M_a->current_NOOP_tag = NULL;

	    DPRINT(Debug,9,(&Debug, 
			    "parse_response: Response for idle timer NOOP command -- ignored.\n"));
	    return 1;
	}
	if (! M_a->current_tag  ||
	    0 != strcmp(M_a->current_tag,
			M_a->read_tokens.tokens[0].str)) {
	    DPRINT(Debug,1,(&Debug, "parse_response: bad tag %s\n",
			    M_a->read_tokens.tokens[0].str));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}

	free(M_a->current_tag);
	M_a->current_tag = NULL;

	if (M_a->command_result_code) {
	    free(M_a->command_result_code);
	    M_a->command_result_code = NULL;
	}

	switch (M_a->read_tokens.tokens[1].imap_token) {
	    struct string *Text;
	    char *code;   /* Do not free() this -- this is just pointer
			     to read_tokens area
			  */	    
	case imap_status_keyword:
	    Text = NULL;
	    code = NULL;
	    pick_text(M_a,&Text,&code);

	    if (code)
		M_a->command_result_code = safe_strdup(code);

	    if (code && Text && 0 == istrcmp(code,"ALERT")) {
		lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapAlert,
				      "IMAP alert: %S"),
			      Text);		
	    } else if (code && 0 == istrcmp(code,"READ-ONLY")) {
		if (M_b) {
		    M_b->folder_status &= ~IMAP_writable;
		    DPRINT(Debug,5,(&Debug, 
				    "parse_response: Folder is read-only\n"));
		} else {
		    DPRINT(Debug,1,(&Debug, 
				    "parse_response: Unexpected READ-ONLY response\n"));
		}
	    } else if (code && 0 == istrcmp(code,"READ-WRITE")) {
		if (M_b) {
		    M_b->folder_status |= IMAP_writable;
		    DPRINT(Debug,5,(&Debug, 
				    "parse_response: Folder is read-write\n"));
		} else {
		    DPRINT(Debug,1,(&Debug, 
				    "parse_response: Unexpected READ-WRITE response\n"));
		}
	    }

	    if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"NO")) {
		M_a->imap_state = IMAP_command_ready_NO;
		if (Text)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapError,
				      "IMAP server: %S"),
			      Text);		
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"BAD")) {
		M_a->imap_state = IMAP_command_ready_BAD;	    
		if (Text)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapError,
				      "IMAP server: %S"),
			      Text);
	    } else if (0 == istrcmp(M_a->read_tokens.tokens[1].str,"OK")) {
		/* Don't forget closing state ... */
		if (IMAP_closing != M_a->imap_state)
		    M_a->imap_state = IMAP_command_ready_OK;	    
	    } else {
		DPRINT(Debug,1,(&Debug, "parse_response: status=%s\n",
				M_a->read_tokens.tokens[1].str));
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
				  "Bad response from IMAP server"));
		if (Text)
		    free_string(&Text);
		return 0;
	    }
	    if (Text)
		free_string(&Text);
	    break;
	default:
	    DPRINT(Debug,1,(&Debug, "parse_response: token[1] type =%d\n",
			    M_a->read_tokens.tokens[1].imap_token));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	break;

    case imap_continue:
	/* TODO: Handle AUTHENTICATE */

	if (M_a->literal_mode != waiting_literal_continue) {
	    DPRINT(Debug,1,(&Debug, 
			    "parse_response: Unexpected continuation request\n"));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	M_a->literal_mode = continue_literal;

	if (!imap_gen_writable_data(M_a)) {
	    panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
		  "Not literal data for writing (IMAP)",0);
	}

	DPRINT(Debug,9,(&Debug, 
			"parse_response=2 (continue literal sending)\n")); 
	return 2;
    default:
	DPRINT(Debug,1,(&Debug, "parse_response: token[0] type =%d\n",
			M_a->read_tokens.tokens[0].imap_token));
	lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			  "Bad response from IMAP server"));
	return 0;
    }

    return 1;
}

static int imap_write_stream P_((struct streamsched *ss,void * data));
static int imap_idle_timer P_((struct streamsched *ss,void * data));

static int imap_read_stream (ss,data)
     struct streamsched *ss; 
     void * data;
{
    struct connection_cache *con = data;
    int n;
    int close_it = 0;
    int tok_ret;
    char * ERROR = NULL;

    if (!data || 
	con->type != &IMAP_connection ||
	con->C.stream != ss) {
	
	panic("MBX PANIC",__FILE__,__LINE__,"imap_read_stream",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug, 
		     "imap_read_stream: con=%p (%s@%s)\n",
		     con,
		     con->C.username ? con->C.username : "<NULL>",
		     con->C.host ? con->C.host : "<NULL>"));

    if (! con->a.imap_con) {
	DPRINT(Debug,13,(&Debug,
                         "imap_read_stream=0 (NO IMAP CONNECTION!)\n"));
	return 0;
    }

    if (IMAP_error == con->a.imap_con->imap_state) {
	DPRINT(Debug,13,(&Debug, 
			 "imap_read_stream=0 (stream need closing)\n"));
	return 0;
    }

    if (con->a.imap_con->wanna_literal)
	n = ReadFromStream(ss, &(con->a.imap_con->read_buffer),
			   con->a.imap_con->literal_len -
			   con->a.imap_con->read_buffer.read_len);
    else
	n = ReadFromStream(ss, &(con->a.imap_con->read_buffer),-1);

    DPRINT(Debug,50,(&Debug, "Buffer content="));
#ifdef DEBUG
    if (Debug.active >= 50) {
	int i;

	for (i = 0; i < con->a.imap_con->read_buffer.read_len; i++) {
	    DPRINT(Debug,50,(&Debug, " %d",
			     con->a.imap_con->read_buffer.read_buffer[i]));
	}
	DPRINT(Debug,50,(&Debug, "\n (message?)="));
	for (i = 0; i < con->a.imap_con->read_buffer.read_len; i++) {
	    DPRINT(Debug,50,(&Debug, " %c",
			     isascii(con->a.imap_con->read_buffer.
				     read_buffer[i]) &&
			     isprint(con->a.imap_con->read_buffer.
				     read_buffer[i]) ? 
			     con->a.imap_con->read_buffer.read_buffer[i] :
			     '.'));
	}
    }
#endif
    DPRINT(Debug,50,(&Debug, "\n"));
    
    /* Parse response ... */
    do {
	tok_ret = tokenize(con->a.imap_con);
	if (tok_ret < 0) {
	    DPRINT(Debug,7,(&Debug, 
			    "imap_read_stream: Failed to tokenize response ...\n"));
	    close_it = 1;
	} else if (tok_ret > 0) {  /* Tokenized */
	    int X;
	    if (!(X = parse_response(con))) {
		DPRINT(Debug,7,(&Debug, 
				"imap_read_stream: Failed to parse response ...\n"));
		close_it = 1;
	    } else if (2 == X) {		
		/* Restart writing of literal ... */

		ConfigStream(con->C.stream,
			     imap_read_stream,imap_write_stream,
			     imap_idle_timer,IMAP_TIMEOUT,con);
		DPRINT(Debug,13,(&Debug, 
				 "imap_read_stream: Continue literal sending...\n"));
	    }
	    
	    free_tokens(& (con->a.imap_con->read_tokens));
	}	
    } while (0  < tok_ret && !close_it &&
	     con->a.imap_con->read_buffer.read_len > 0);
    
    if (n == 0) {
	if (con->a.imap_con->imap_state != IMAP_closing)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectionClosedImap,
			      "Connection closed to IMAP server"));
	close_it = 1;
    }

    if (n < 0) {
	ERROR = RemoveStreamError(ss);
	if (ERROR) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeErrorReadingImap,
			      "Error reading from IMAP server: %s"),
		      ERROR);
	    close_it = 1;
	}
    }

    if (ERROR) {
	free(ERROR);
	ERROR = NULL;
    }

    if (close_it) {
	con->a.imap_con->imap_state = IMAP_error;
	DPRINT(Debug,9,(&Debug, 
			"imap_read_stream=0 (stream need closing)\n"));
	return 0;
    }
    DPRINT(Debug,13,(&Debug, 
		     "imap_read_stream=1\n"));
    return 1;
}


static int imap_write_stream (ss,data)
     struct streamsched *ss;
     void * data;
{
    struct connection_cache *con = data;
    char * ERROR = NULL;
    int n;


    if (!data || 
	con->type != &IMAP_connection ||
	con->C.stream != ss) {
	
	panic("MBX PANIC",__FILE__,__LINE__,"imap_write_stream",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug, 
		"imap_write_stream: folder=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));

    if (!con->a.imap_con) {
	DPRINT(Debug,7,(&Debug,
                        "imap_write_stream=0 (NO imap connection)\n"));
	return 0;
    }


    if (!imap_gen_writable_data(con->a.imap_con)) {
	DPRINT(Debug,7,(&Debug, 
			"imap_write_stream: NO DATA TO WRITE!\n"));
    }

    if (IMAP_error == con->a.imap_con->imap_state) {
	DPRINT(Debug,7,(&Debug, 
			"imap_write_stream=0 (stream need closing)\n"));
	return 0;
    }

    n = WriteToStream(ss,& (con->a.imap_con->write_buffer));


    if (n > 0) {
	int i;
	
	DPRINT(Debug,70,(&Debug, "Write content="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug, " %d",
			con->a.imap_con->write_buffer.write_buffer[i]));
	}
	DPRINT(Debug,70,(&Debug, "\n    (text?)="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug, " %c",
			     isascii(con->a.imap_con->write_buffer.write_buffer[i]) 
			     &&
			     isprint(con->a.imap_con->write_buffer.write_buffer[i]) 
			     ? 
			     con->a.imap_con->write_buffer.write_buffer[i] :
			     '.'));
	}
	DPRINT(Debug,70,(&Debug, "\n"));

	cut_Write_Buffer(& (con->a.imap_con->write_buffer),n);
	
    } else if (n < 0) {
	ERROR = RemoveStreamError(ss);

	DPRINT(Debug,7,(&Debug, 
			"imap_write_stream: write error %s\n",
			ERROR ? ERROR : "(none)"));
    }

    if (!con->a.imap_con->write_buffer.write_len) {
	if (!imap_gen_writable_data(con->a.imap_con)) {
	    if (ERROR) {
		free(ERROR);
		ERROR = NULL;
	    }
	    DPRINT(Debug,13,(&Debug, 
			     "imap_write_stream=0 -- data written\n"));

	    return 0;
	}
    }

    if (ERROR) {
	lib_error(CATGETS(elm_msg_cat, MeSet,MeErrorWritingImap,
			  "Error writing to IMAP server: %s"),
		  ERROR);
	con->a.imap_con->imap_state = IMAP_error;
	DPRINT(Debug,7,(&Debug, 
			"imap_write_stream=0 (stream need closing)\n"));
	free(ERROR);
	ERROR = NULL;
        return 0;	
    }

    DPRINT(Debug,13,(&Debug, "imap_write_stream=1\n"));
    return 1;
}

#define MAX_TAG 99999

static char * tag_generator P_((void));
static char * tag_generator()
{
    /* NOTE: There is no need to have per imap connection tag 
       generating sequence
    */ 
    static char tag_letter = 'A';
    static int  tag_number = 0;

    tag_number++;
    if (tag_number > MAX_TAG) {
	tag_number = 1;
	tag_letter++;
	if ('Z' == tag_letter) {
	    tag_letter = 'A';
	    DPRINT(Debug,13,(&Debug, 
			     "tag_generator: Tag wrapping around\n"));
	}
    }

    return elm_message(FRM("%c%05d"),tag_letter,tag_number);
}

static int imap_idle_timer (ss,data)
     struct streamsched *ss;
     void * data;
{
    struct connection_cache *con = data;

    if (!data || 
	con->type != &IMAP_connection ||
	con->C.stream != ss) {

	panic("MBX PANIC",__FILE__,__LINE__,"imap_idle_timer",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug, 
		"imap_idle_timer: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));

    if (IMAP_error == con->a.imap_con->imap_state) {
	DPRINT(Debug,7,(&Debug, 
			"imap_idle_timer=0 (stream need closing)\n"));
	return 0;
    }

    if (con->a.imap_con->current_NOOP_tag) {
	DPRINT(Debug,13,(&Debug, 
			 "imap_idle_timer=1 (previous NOOP command is not finished)\n"));
	return 1;
    }

    if (con->a.imap_con->token_index >= 0 || 
	con->a.imap_con->imap_state != IMAP_idle) {
	DPRINT(Debug,13,(&Debug, 
			 "imap_idle_timer=1 (NOT idle)\n"));
	return 1;	
    }
    con->a.imap_con->current_NOOP_tag = tag_generator();
    con->a.imap_con->token_index      = -2;    /* Produce NOOP command */
    con->a.imap_con->imap_state       = IMAP_command;   /* For imap_gen_writable_data */

    if (!imap_gen_writable_data(con->a.imap_con)) {
	panic("MBX PANIC",__FILE__,__LINE__,"imap_idle_timer",
	      "No NOOP command for writing (IMAP)",0);
    }

    ConfigStream(con->C.stream,imap_read_stream,imap_write_stream,
		 imap_idle_timer,IMAP_TIMEOUT,con);

    /* Mark it again IDLE ... */
    con->a.imap_con->imap_state = IMAP_idle;

    return 1;
}

void imap_wait(con)
     struct connection_cache *con;
{
    if (!con->a.imap_con) {
	DPRINT(Debug,7,(&Debug,"imap_wait: No imap connection\n"));
	return;
    }

    while (IMAP_command_ready_OK  != con->a.imap_con->imap_state &&
	   IMAP_command_ready_NO  != con->a.imap_con->imap_state &&
	   IMAP_command_ready_BAD != con->a.imap_con->imap_state &&
	   IMAP_closing           != con->a.imap_con->imap_state &&
	   IMAP_error             != con->a.imap_con->imap_state &&
           NULL                   != con->C.stream) {

	WaitStreamFor(con->C.stream,SS_read_act);
	/* TODO:    Should user to have some way to interrupt progress ? */
    }
}

void imapcon_idle_wait(con)
     struct connection_cache *con;
{

   struct IMAP_CON * Ma;

    if (con->type != &IMAP_connection) {
	panic("MBX PANIC",__FILE__,__LINE__,"imapcon_idle_wait",
	      "Wrong type connection attached to folder",0);
    }
   
    Ma = con->a.imap_con;

    if (!Ma)
	panic("MBX PANIC",__FILE__,__LINE__,"imapcon_idle_wait",
              "No imap connection",0);

    if (Ma->current_NOOP_tag &&
	Ma->imap_state != IMAP_closing &&
	Ma->imap_state != IMAP_error) {
	DPRINT(Debug,3,(&Debug, 
			"imapcon_idle_wait: Waiting for idle NOOP complete...\n"));
	while(Ma->current_NOOP_tag && 
	      Ma->imap_state != IMAP_closing &&
	      Ma->imap_state != IMAP_error &&
	      con->C.stream) {
	    WaitStreamFor(con->C.stream,SS_read_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	
	DPRINT(Debug,3,(&Debug, 
		    "imapcon_idle_wait: idle NOOP wait completed\n"));
    }
    if (IMAP_idle != Ma->imap_state) {
	DPRINT(Debug,3,(&Debug, 
		    "imapcon_idle_wait: Waiting for command complete...\n"));

	imap_wait(con);
	DPRINT(Debug,3,(&Debug, 
			"imapcon_idle_wait: command wait completed\n"));
    }

}

int imap_command_ok(con,cmd_result,extra_code) 
     struct connection_cache *con;
     imap_states *cmd_result;
     char **extra_code;
{
    int status = 0;
    DPRINT(Debug,12,(&Debug, 
		"imap_command_ok: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));


    if (extra_code) 
	*extra_code = NULL;

    if (IMAP_idle == con->a.imap_con->imap_state) {
	DPRINT(Debug,7,(&Debug, 
		    "imap_command_ok: WRONG STATE\n"));
	status = 0;
	*cmd_result = IMAP_error;
	goto clean;
    }
    imap_wait(con);

    status = IMAP_command_ready_OK == con->a.imap_con->imap_state;
    *cmd_result = con->a.imap_con->imap_state;

    if (IMAP_command_ready_OK == con->a.imap_con->imap_state ||
	IMAP_command_ready_NO == con->a.imap_con->imap_state ||
	IMAP_command_ready_BAD == con->a.imap_con->imap_state) {
	
	if (extra_code) {
	    *extra_code = con->a.imap_con->command_result_code;
	    con->a.imap_con->command_result_code = NULL;	    
	}
    }

    if (IMAP_error == con->a.imap_con->imap_state) {
	if (con->C.stream) {
	    DPRINT(Debug,7,(&Debug, 
			    "imap_command_ok: Closing stream\n"));
	    
	    FreeStreamStack ( &(con->C.stream));
	}	
	status = 0;
	goto clean;
    }

 clean:
    DPRINT(Debug,12,(&Debug, 
		     "imap_command_ok=%d\n",
		     status));
    return status;
}


int start_imap_command(con,verb)
     struct connection_cache *con;
     char * verb;
{
    int status = 0;

    DPRINT(Debug,12,(&Debug, 
		"start_imap_command: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));

    if (IMAP_closing == con->a.imap_con->imap_state) {
	DPRINT(Debug,12,(&Debug, 
			 "start_imap_command: imap_state == IMAP_closing\n"));
	goto failure;
    }

    if (IMAP_error == con->a.imap_con->imap_state) {
	DPRINT(Debug,12,(&Debug, 
			 "start_imap_command: imap_state == IMAP_error\n"));

    failure:
	if (con->C.stream ) {
	    DPRINT(Debug,12,(&Debug, 
			"start_imap_command: Closing stream\n"));
	    
	    FreeStreamStack(& (con->C.stream));
	}	
	status = 0;
	goto clean;
    }

    if (IMAP_idle != con->a.imap_con->imap_state) {
	DPRINT(Debug,7,(&Debug, 
		    "start_imap_command: wrong state  ... waiting...\n"));
	
	imap_wait(con);
	DPRINT(Debug,7,(&Debug, 
		    "start_imap_command: ... wait ended\n"));
	if (IMAP_error == con->a.imap_con->imap_state) {
	    DPRINT(Debug,12,(&Debug, 
			     "start_imap_command: imap_state == IMAP_error\n"));

	    goto failure;	
	}

	if (IMAP_closing == con->a.imap_con->imap_state) {
	    DPRINT(Debug,12,(&Debug, 
			     "start_imap_command: imap_state == IMAP_closing\n"));
	    goto failure;
	}
    }
    
    if (con->a.imap_con->current_tag) {
	DPRINT(Debug,12,(&Debug, 
			 "start_imap_command: Tag %s still active\n",
			 con->a.imap_con->current_tag));
	status = 0;
	goto clean;	
    }

    con->a.imap_con->imap_state  = IMAP_command;
    free_tokens(& (con->a.imap_con->write_tokens));
    con->a.imap_con->token_index  = -1;
    con->a.imap_con->literal_mode = no_literal;

    if (NULL == con->C.stream) {
	DPRINT(Debug,12,(&Debug, 
		    "start_imap_command: stream not open\n"));
	status = 0;
	goto clean;
    }

    con->a.imap_con->current_tag = tag_generator();
    imap_command_push_atom(con,verb);
    status          = 1; 

    DPRINT(Debug,15,(&Debug, 
		"start_imap_command: Command %s ...\n",verb));

 clean:
    DPRINT(Debug,12,(&Debug, 
		"start_imap_command=%d\n",
		status));
    return status;    
}


void imap_command_push_atom(con,st)
     struct connection_cache *con;
     const char *st;
{
    struct imap_token T;

    T.imap_token    = imap_atom;
    T.str           = safe_strdup(st);
    T.len_or_value  = strlen(st);
    add_token(&(con->a.imap_con->write_tokens),T);
}

void imap_command_push_literal(con,len,st)
     struct connection_cache *con;
     int len;
     const char *st;
{
    struct imap_token T;

    if (len < 0) {
	panic("MBX PANIC",__FILE__,__LINE__,
	      "imap_command_push_literal","negative len",0);
    }

    T.imap_token    = imap_literal;
    T.str           = safe_malloc(len+1);
    memcpy(T.str,st,len);
    T.str[len]      = '\0';
    T.len_or_value  = len;
    add_token(&(con->a.imap_con->write_tokens),T);
}


void imap_command_push_string(con,st)
     struct connection_cache *con;
     const char *st;
{
    struct imap_token T;

    T.imap_token    = imap_string;
    T.str           = safe_strdup(st);
    T.len_or_value  = strlen(st);
    add_token(&(con->a.imap_con->write_tokens),T);
}

void imap_command_push_astring(con,st)
     struct connection_cache *con;
     const char *st;
{
    if ('\0' == st[0] ||
	strspn(st,"ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvxyz0123456789")
	!= strlen(st))
	imap_command_push_string(con,st);
    else 
	imap_command_push_atom(con,st);
}


void imap_command_push_aliteral(con,st)
     struct connection_cache *con;
     const char *st;
{
    int len = strlen(st);

    if ('\0' == st[0] ||
	strspn(st,"ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvxyz0123456789")
	!= len)
	imap_command_push_literal(con,len,st);
    else 
	imap_command_push_atom(con,st);
}


void imap_command_push_range(con,A,B)
     struct connection_cache *con;
     int A,B;
{
    char * st = elm_message(FRM("%d:%d"),A,B);
    imap_command_push_atom(con,st);
    free(st);
}


void imap_command_push_number(con,A)
     struct connection_cache *con;
     int A;
{
    char * st = elm_message(FRM("%d"),A);
    imap_command_push_atom(con,st);
    free(st);
}

void imap_command_push_list(con,l)
     struct connection_cache *con;
     const char **l;
{
    int i;
    struct imap_token T;

    T.imap_token    = imap_list_begin;
    T.str           = NULL;
    T.len_or_value  = 0;
    add_token(&(con->a.imap_con->write_tokens),T);

    for (i = 0; l[i]; l++)
	imap_command_push_atom(con,l[i]);

    T.imap_token    = imap_list_end;
    T.str           = NULL;
    T.len_or_value  = 0;
    add_token(&(con->a.imap_con->write_tokens),T);
}

void end_imap_command(con)
     struct connection_cache *con;
{
    DPRINT(Debug,12,(&Debug, 
		"end_imap_command: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));

    if (IMAP_error == con->a.imap_con->imap_state) {
	DPRINT(Debug,12,(&Debug, 
			 "end_imap_command:  On error state\n"));
	return;
    }

    if (IMAP_closing == con->a.imap_con->imap_state) {
	DPRINT(Debug,12,(&Debug, 
			 "end_imap_command:  Is closing\n"));
	return;
    }

    if (con->a.imap_con->imap_state != IMAP_command)
	panic("MBX PANIC",__FILE__,__LINE__,
	      "end_imap_command","wrong state",0);

    if (!imap_gen_writable_data(con->a.imap_con)) {
	panic("MBX PANIC",__FILE__,__LINE__,"end_imap_command",
	      "No command for writing",0);
    }
    ConfigStream(con->C.stream,imap_read_stream,imap_write_stream,
		 imap_idle_timer,IMAP_TIMEOUT,con);
}

/* ----------------------------------------------------------------------- */

/* IMAP connection */

static void  cache_zero_imap(con)
     struct connection_cache *con;
{
    DPRINT(Debug,11,(&Debug, "cache_zero_imap: con=%p\n",con));

    con->a.imap_con = safe_malloc (sizeof (* (con->a.imap_con)));
    bzero((void *)con->a.imap_con, sizeof (* (con->a.imap_con)));

    con->a.imap_con->capability_bits      = 0;
#ifdef USE_DLOPEN
    con->a.imap_con->capability_libs      = NULL;
    con->a.imap_con->capability_lib_count = 0;
#endif

    zero_Read_Buffer(&(con->a.imap_con->read_buffer));
    con->a.imap_con->wanna_literal   = 0;
    con->a.imap_con->literal_len     = 0;

    con->a.imap_con->read_tokens.tokens          = NULL;
    con->a.imap_con->read_tokens.token_count     = 0;
    con->a.imap_con->tokenizer_state             = state_tag;

    con->a.imap_con->current_tag         = NULL;
    con->a.imap_con->imap_state          = IMAP_idle;
    con->a.imap_con->command_result_code = NULL;

    con->a.imap_con->write_tokens.tokens      = NULL;
    con->a.imap_con->write_tokens.token_count = 0;
    con->a.imap_con->token_index  = -1;

    zero_Write_Buffer(& (con->a.imap_con->write_buffer));
    con->a.imap_con->literal_mode = no_literal;

    con->a.imap_con->current_NOOP_tag = NULL;
}

static void cache_imap_clear_buffers P_((struct connection_cache *con));
static void cache_imap_clear_buffers(con)
     struct connection_cache *con;
{
    DPRINT(Debug,12,(&Debug, 
		"cache_imap_clear_buffers: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));
        
    con->a.imap_con->capability_bits = 0;
#ifdef USE_DLOPEN
    free_only_imap_capa_libs( & (con->a.imap_con->capability_libs),
			      & (con->a.imap_con->capability_lib_count));
#endif

    free_Read_Buffer(&(con->a.imap_con->read_buffer));

    con->a.imap_con->wanna_literal = 0;
    con->a.imap_con->literal_len   = 0;
    
    free_tokens(&(con->a.imap_con->read_tokens));

    con->a.imap_con->tokenizer_state = state_tag;

    if (con->a.imap_con->current_tag) {
	free(con->a.imap_con->current_tag);
	con->a.imap_con->current_tag = NULL;
    }
    con->a.imap_con->imap_state = IMAP_error;
    if (con->a.imap_con->command_result_code) {
	free(con->a.imap_con->command_result_code);
	con->a.imap_con->command_result_code = NULL;
    }

    free_tokens(&(con->a.imap_con->write_tokens));
    con->a.imap_con->token_index = -1;
    
    
    free_Write_Buffer ( &(con->a.imap_con->write_buffer));
    con->a.imap_con->literal_mode = no_literal;

    if (con->a.imap_con->current_NOOP_tag) {
	free(con->a.imap_con->current_NOOP_tag);
	con->a.imap_con->current_NOOP_tag = NULL;
    }
}

#ifdef USE_DLOPEN
static struct imap_callbacks imap_commands = {
    imap_wait,
    imapcon_idle_wait,
    imap_command_ok,
    start_imap_command,
    imap_command_push_atom,
    imap_command_push_literal,
    imap_command_push_string,
    imap_command_push_astring,
    imap_command_push_aliteral,
    imap_command_push_range,
    imap_command_push_number,
    imap_command_push_list,
    end_imap_command,
    imap_clear_command
};
#endif

static void  cache_free_imap(con)
     struct connection_cache *con;
{    
    DPRINT(Debug,11,(&Debug, 
		     "cache_free_imap: con=%p (%s@%s)\n",
		     con,
		     con->C.username ? con->C.username : "<NULL>",
		     con->C.host ? con->C.host : "<NULL>"));
    
    if (con->a.imap_con) {
	cache_imap_clear_buffers(con);

	bzero((void *)con->a.imap_con, sizeof (* (con->a.imap_con)));

	free(con->a.imap_con);
	con->a.imap_con = NULL;
    }

}

static int  cache_open_imap(con)
     struct connection_cache *con;
{
    int status = 0;
    imap_states res;

    DPRINT(Debug,11,(&Debug, 
		     "cache_open_imap: con=%p (%s@%s)\n",
		     con,
		     con->C.username ? con->C.username : "<NULL>",
		     con->C.host ? con->C.host : "<NULL>"));
    
    /* Clear buffer from old data ... */
    cache_imap_clear_buffers(con);
    
    if (NULL == con->C.stream || 
	con->state == CON_error) {
	struct service_entry *se = 
	    give_service_entry(con->C.host,STFLAG_is_imap);
	PORTS IMAP_ports[] = { PORT_imap4, PORT_end };
	PORTS force_port = PORT_end;

	int got;
	
	if (!se) {
	    status = 0;
	    con->state = CON_error;
	    goto clean;
	}

	free(con->C.host);
	con->C.host = safe_strdup(se->official_name);

	if (con->port) {
	    force_port = con->port;
	    DPRINT(Debug,7,(&Debug, 
			    "cache_open_imap: force_port = %d\n"));
	}

	if (!connect_remote_account(&(con->C),
				    &got,se,
				    IMAP_ports,
				    force_port)) {
	    status = 0;
	    con->state = CON_error;
	    free_temporary_service_entry(&se);
	    goto clean;
	}
	free_temporary_service_entry(&se);
	con->state = CON_greeting;
    }
    
    if (con->state == CON_greeting) {

	DPRINT(Debug,7,(&Debug, 
			"cache_open_imap: Waiting for server greeting...\n"));

	/* We want server's initial response to connection ...
	   .. maybe untagged BYE?
	*/
	con->a.imap_con->imap_state = IMAP_idle;
	ConfigStream(con->C.stream,
		     imap_read_stream,ss_noaction_routine,imap_idle_timer,
		     IMAP_TIMEOUT,con);
	WaitStreamFor(con->C.stream,SS_read_act);
	if (NULL ==  con->C.stream ||
	    IMAP_error   == con->a.imap_con->imap_state ||
	    IMAP_closing == con->a.imap_con->imap_state) {
	    DPRINT(Debug,7,(&Debug, 
			    "cache_open_imap: Server does not accept connection\n"));
	    con->state = CON_error;
	    goto clean;
	}
	con->state = CON_open;
    }
    /* Assume that everything OK */
    status = 1;
    con->a.imap_con->imap_state = IMAP_idle;

#ifdef USE_DLOPEN
 retry_prelogin_capa:
#endif

    if (!start_imap_command(con,"CAPABILITY")) 
	goto clean;
    end_imap_command(con);
    
    if (!imap_command_ok(con,&res,NULL))
	goto clean;

#ifdef USE_DLOPEN
    else {
	enum CAPA_phase phase = capa_prelogin;
	
	DPRINT(Debug,13,(&Debug, "%d libraries available\n",
		    con->a.imap_con->capability_lib_count));
	if (!handle_imap_capa_libs(con,
				   & (con->a.imap_con->capability_libs),
				   & (con->a.imap_con->capability_lib_count),
				   &phase, &imap_commands)) {
	    status = 0;
	    con->state = CON_error;
	}

	switch(phase) {
	case capa_prelogin_again:
	    goto retry_prelogin_capa;
	case capa_logged:
	    con->state = CON_logged;
	    break;
	default:
	    break;
	}

    }
#endif

 clean:
    imap_clear_command(con);

    DPRINT(Debug,11,(&Debug, 
		     "cache_open_imap=%d\n",status)); 
    return status;
}

static int  cache_login_imap(con,password)
     struct connection_cache *con;
     const struct string *password; /* May be NULL */
{
    char * data1 = NULL;
    imap_states res;
    int status = 0;
    int bits;

    DPRINT(Debug,11,(&Debug, 
		     "cache_login_imap: con=%p (%s@%s)\n",
		     con,
		     con->C.username ? con->C.username : "<NULL>",
		     con->C.host ? con->C.host : "<NULL>"));

    if (con->state != CON_open ||
	NULL == con->C.stream) {
	DPRINT(Debug,11,(&Debug, 
			 "cache_login_imap: state not open\n")); 
	con->state = CON_error;
	status = 0;
	goto clean;
    }

    StreamInfo(con -> C.stream,SS_ssf,&bits,NULL,NULL);


    if (!start_imap_command(con,"LOGIN")) 
	goto clean;

    imap_command_push_astring(con,con->C.username);

    if (bits < 2)
	lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginClear,
			      "IMAP login to %s as %s ... (clear text)"),
		      con->C.host,
		      con->C.username);    

    else {
	int verify_status = 0;
	struct string * peer_name = NULL;
	
	StreamInfo(con -> C.stream,SS_verify,&verify_status,NULL,
		   &peer_name);
	
	if (verify_status < 0)
	    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginVerFail,
				  "IMAP login to %s as %s ... (VERIFY FAILED)"),
			  con->C.host,
			  con->C.username);

	else if (verify_status > 0)  {

	    if (peer_name) {
		/* XXX Is charset correct? */
		struct string * temp = new_string2(system_charset,
						   s2us(con->C.host));

		/* Optimize display */

		if (0 == string_cmp(peer_name,temp,99)) 
		    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginVer,
					  "IMAP login to %S (verified) as %s ..."),
				  peer_name,
				  con->C.username);
		else
		    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginVerAs,
					  "IMAP login to %s as %s ... (verified: %S)"),
				  con->C.host,
				  con->C.username,
				  peer_name);
		
		free_string(&temp);

	    } else
		lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLogin,
				      "IMAP login to %s as %s ..."),
			      con->C.host,
			      con->C.username);

	    /* If no name, verify result is not very usefull */

	} else 
	    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginVerNot,
				  "IMAP login to %s as %s ... (not verified)"),
			  con->C.host,
			  con->C.username);

	if (peer_name)
	    free_string(&peer_name);

    }

    if (password) {
	struct string * pass1;

	if (!  can_ascii_string(password)) {
	    DPRINT(Debug,1,(&Debug, 
			    "cache_login_imap: Non ascii password are not supported\n"));
	}

	/* NOTE: Original string is kept, if it can not converted
	   to US-ASCII -- then this produces garbage
	*/
	pass1 = ascify_string(password);

	data1 =  us2s(stream_from_string(pass1,0,NULL));

	free_string(&pass1);

    } else
	data1 = lib_prompt(1,CATGETS(elm_msg_cat, MeSet,MeLoginPasswd,
				     "Password for %s@%s:"),
			   con->C.username,
			   con->C.host);
    if (!data1) {
	status = 0;
	goto clean;	    
    }

    /* Put is as literal
    */
    imap_command_push_aliteral(con,data1);
    end_imap_command(con);

    if (!imap_command_ok(con,&res,NULL)) {

	/* Show possible login error */

	if (sleepmsg > 0) {  /* HACK */

	    /* POLL_METHOD is defined because 
	       REMOTE_MBX is defined */

	    wait_for_timeout(sleepmsg);

	}
	goto clean;
    }
    con->state = CON_logged;
    status = 1;

    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLoginOK,
			  "IMAP login to %s as %s ... OK"),
		  con->C.host,
		  con->C.username);
    
 clean:
    imap_clear_command(con);

    if (data1) {
	free(data1);
	data1 = NULL;
    }
    DPRINT(Debug,11,(&Debug, 
		     "cache_login_imap=%d\n",status)); 
    
    return status;
}


static int  cache_close_imap(con)
     struct connection_cache *con;
{
    DPRINT(Debug,12,(&Debug, 
		     "cache_close_imap: con=%p (%s@%s)\n",
		     con,
		     con->C.username ? con->C.username : "<NULL>",
		     con->C.host ? con->C.host : "<NULL>"));
    
    if (con->state != CON_error) {
	DPRINT(Debug,12,(&Debug, 
			 "cache_close_imap: LOGOUT\n"));
	
	if (start_imap_command(con,"LOGOUT")) {
	    imap_states res;
	    
	    end_imap_command(con);
	    imap_command_ok(con,&res,NULL);
	    imap_clear_command(con);
	}
    }

    if (con->C.stream ) {
	DPRINT(Debug,12,(&Debug, 
			 "cache_close_imap: Closing stream\n"));
	
	FreeStreamStack( &(con->C.stream));
    }	

    con->state = CON_error;
    return 1;
}


static void cache_folder_from_imap(con,folder)
     struct connection_cache *con;
     struct folder_info *folder;
{
    DPRINT(Debug,12,(&Debug, 
		"cache_folder_from_imap: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));
    
    if (folder->folder_type != &imap_mbx) 
	panic("MBX PANIC",__FILE__,__LINE__,"cache_folder_from_imap",
	      "Can't attach folder type to connection",0);

    if (folder->p->a.imap_mbx.Ch &&
	folder->p->a.imap_mbx.Ch != con) {
	DPRINT(Debug,12,(&Debug, 
		    "cache_folder_from_imap: Freeing old connection: %p\n",
		    folder->p->a.imap_mbx.Ch));
	free_connection(&(folder->p->a.imap_mbx.Ch));
    }

    folder->p->a.imap_mbx.Ch            = con;
    folder->p->a.imap_mbx.folder_status = 0;

    /* TODO :  Check necessary actions */
    
}

static void cache_browser_from_imap(con,dir)
     struct connection_cache *con;
     struct folder_browser   *dir;
{
    DPRINT(Debug,12,(&Debug, 
		"cache_browser_from_imap: con=%p (%s@%s)\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>"));
    
    if (dir->type != &imap_browser) 
	panic("MBX PANIC",__FILE__,__LINE__,"cache_browser_from_imap",
	      "Can't attach browser type to connection",0);

    DPRINT(Debug,12,(&Debug, " ... caching %s\n",
		     IMAP_connection_cache ? "enabled" : "disabled"));

    if (IMAP_BROWSER_magic != dir->a.imap_browser->magic)
	panic("MBX PANIC",__FILE__,__LINE__,"cache_browser_from_imap",
	      "Bad magic type",0);


    if (dir->a.imap_browser->Ch &&
	dir->a.imap_browser->Ch != con) {

	if (dir->a.imap_browser->Ch->state != CON_error &&
	    IMAP_connection_cache) {
	    DPRINT(Debug,12,(&Debug, 
			"cache_browser_from_imap: Caching old connection: %p\n",
			dir->a.imap_browser->Ch));
	    cache_connection(dir->a.imap_browser->Ch);
	} else {
	    DPRINT(Debug,12,(&Debug, 
			     "cache_browser_from_imap:  freeing old connection\n"));
	    free_connection(&(dir->a.imap_browser->Ch));
	}
    }

    dir->a.imap_browser->Ch            = con;

    /* TODO :  Check necessary actions */
    
}

#endif /* REMOTE_MBX */



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
